# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import logging

import fastText
import numpy as np

from elit.embedding import TokenEmbedding

__author__ = "Gary Lai"


class FastText(TokenEmbedding):
    """
    :class:`FastText` is a token-based model trained by `FastText <https://github.com/facebookresearch/fastText>`_.
    """

    def __init__(self, filepath: str):
        """
        :param filepath: the path to the binary file containing a word embedding model trained by FastText (``*.bin``).
        """
        logging.info('FastText')
        logging.info('- model: {}'.format(filepath))
        self.model = fastText.load_model(filepath)
        dim = self.model.get_dimension()
        super().__init__(dim)
        logging.info('- vocab = %d, dim = %d' % (len(self.model.get_words()), dim))

    # override
    def emb(self, value: str) -> np.ndarray:
        return self.model.get_word_vector(value)



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from typing import Sequence

import numpy as np

from elit.structure import Document

__author__ = "Jinho D. Choi, Gary Lai"


class Embedding(abc.ABC):
    """
    :class:`Embedding` is an abstract class to implement embedding models.

    :ivar dim: the dimension of each embedding.
    :ivar pad: the zero vector whose dimension is the same as the embedding (can be used for zero-padding).

    Abstract methods to be implemented:
      - :meth:`Embedding.embed`
    """

    def __init__(self, dim: int):
        """
        :param dim: the dimension of each embedding.
        """
        self.dim = dim
        self.pad = np.zeros(dim).astype('float32')

    @abc.abstractmethod
    def embed(self, docs: Sequence[Document], key: str, **kwargs):
        """
        Adds embeddings to the input documents.

        :param docs: a sequence of input documents.
        :param key: the key to a sentence or a document where embeddings are to be added.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import hashlib
import json
import logging
import os
import pathlib
import re
import sys
from typing import Set

import requests
from tqdm import tqdm

__author__ = "Jinho D. Choi, Gary Lai"


def pkl(filepath):
    return filepath + '.pkl'


def gln(filepath):
    return filepath + '.gln'


def params(filepath):
    return filepath + '.params'


def read_word_set(filename) -> Set[str]:
    """
    :param filename: the name of the file containing one key per line.
    :return: a set containing all keys in the file.
    """
    with open(filename, encoding='utf-8') as fin:
        s = set(line.strip() for line in fin)
    logging.info('Init: %s (keys = %d)' % (filename, len(s)))
    return s


def read_concat_word_dict(filename) -> dict:
    """
    :param filename: the name of the file containing one key per line.
    :return: a dictionary whose key is the concatenated word and value is the list of split points.
    """
    def key_value(line):
        l = [i for i, c in enumerate(line) if c == ' ']
        l = [i - o for o, i in enumerate(l)]
        line = line.replace(' ', '')
        l.append(len(line))
        return line, l

    with open(filename, encoding='utf-8') as fin:
        d = dict(key_value(line.strip()) for line in fin)
    logging.info('Init: %s (keys = %d)' % (filename, len(d)))
    return d


class NoIndent(object):
    def __init__(self, value):
        self.value = value


class NoIndentEncoder(json.JSONEncoder):
    REGEX = re.compile(r'@@@(\d+)@@@')

    def __init__(self, *args, **kwargs):
        super(NoIndentEncoder, self).__init__(*args, **kwargs)
        self.kwargs = dict(kwargs)
        del self.kwargs['indent']
        self._replacements = {}

    def default(self, o):
        if isinstance(o, NoIndent):
            key = len(self._replacements)
            self._replacements[key] = json.dumps(o.value, **self.kwargs)
            return "@@@%d@@@" % key
        else:
            return super(NoIndentEncoder, self).default(o)

    def encode(self, o):
        result = super(NoIndentEncoder, self).encode(o)
        out = []
        m = self.REGEX.search(result)
        while m:
            key = int(m.group(1))
            out.append(result[:m.start(0) - 1])
            out.append(self._replacements[key])
            result = result[m.end(0) + 1:]
            m = self.REGEX.search(result)
        return ''.join(out)


def check_resource_dir(directory):
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)


def file_exist(filename) -> bool:
    return os.path.isfile(filename)


def remove_file(filename):
    os.remove(filename)


def sha1sum(filename):
    buf_size = 65536
    sha1 = hashlib.sha1()
    with open(filename, 'rb') as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def download(source, filename):
    try:
        with open(filename, "wb") as f:
            response = requests.get(source, stream=True)
            total_size = response.headers.get('content-length', 0)

            if total_size is None:
                f.write(response.content)
            else:
                block_size = 32 * 1024
                total_size = int(total_size) // block_size
                with tqdm(total=total_size, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
                    for data in response.iter_content(chunk_size=block_size):
                        f.write(data)
                        pbar.set_description('Downloading: {}'.format(filename))
                        pbar.update(len(data))
    except KeyboardInterrupt:
        if file_exist(filename):
           os.remove(filename)
        sys.exit()


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import argparse
import logging
import sys

__author__ = "Gary Lai"


class BaseCLI(abc.ABC):

    def __init__(self, name: str, usage: str = None, description: str = None):
        """
        :param name: the name of the component.
        :param description: the description of this component; if ``None``, the name is used instead.
        """
        if not description:
            description = name

        parser = argparse.ArgumentParser(usage=usage, description=description)
        parser.add_argument('command', help='{} command to run'.format(name))
        args = parser.parse_args(sys.argv[2:3])

        if not hasattr(self, args.command):
            logging.info('Unrecognized command: ' + args.command)
            parser.print_help()
            exit(1)
        getattr(self, args.command)()


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import Sequence, Optional, List

from mxnet import gluon
from types import SimpleNamespace

__author__ = "Gary Lai"


class CNNModel(gluon.HybridBlock):
    """
    :class:`NLPModel` is an abstract class providing helper methods to implement a neural network model.
    """

    def __init__(self, input_config: SimpleNamespace, output_config: SimpleNamespace,
                 fuse_conv_config: SimpleNamespace=None, ngram_conv_config: SimpleNamespace=None, hidden_configs: SimpleNamespace=None):
        super().__init__()
        self.input_layer = self._init_input_layer(input_config)
        self.output_layer = self._init_output_layer(output_config)
        self.fuse_conv_layer = self._init_fuse_conv_layer(fuse_conv_config)
        self.ngram_conv_layers = self._init_ngram_conv_layers(ngram_conv_config)
        self.hidden_layers = self._init_hidden_layers(hidden_configs)

    def hybrid_forward(self, F, x, *args, **kwargs):
        # x: batch_size, window size, features

        # input layer
        x = self.input_layer.dropout(x)

        # dimensionality reduction layer
        if self.fuse_conv_layer is not None:
            x = F.reshape(x, (0, 1, self.input_layer.row, self.input_layer.col))
            x = self.fuse_conv_layer.dropout(self.fuse_conv_layer.conv(x))

        # convolution layer
        if self.ngram_conv_layers is not None:
            if self.fuse_conv_layer is not None:
                x = F.transpose(x, (0, 3, 2, 1))
            else:
                x = F.reshape(x, (0, 1, self.input_layer.row, self.input_layer.col))
            c = [layer.dropout(layer.pool(layer.conv(x))) if layer.pool else layer.dropout(layer.conv(x).reshape((0, -1))) for layer in self.ngram_conv_layers]
            x = F.concat(*c, dim=1)

        # hidden layers
        if self.hidden_layers is not None:
            for layer in self.hidden_layers:
                x = layer.dense(x)
                x = layer.dropout(x)

        # output layer
        x = self.output_layer.dense(x)
        return x

    # ======================================== create_layers ========================================

    def _init_input_layer(self, config: SimpleNamespace) -> SimpleNamespace:
        """
        :param config: the output of :meth:`FFNNModel.namespace_input_layer`.
        :return: the namespace of (dropout) for the input layer.
        """
        layer = SimpleNamespace(
            col=config.col,
            row=config.row,
            dropout=gluon.nn.Dropout(config.dropout)
        )

        with self.name_scope():
            self.__setattr__(layer.dropout.name, layer.dropout)

        return layer

    def _init_output_layer(self, config: SimpleNamespace) -> SimpleNamespace:
        """
        :param config: the output of :meth:`FFNNModel.namespace_output_layer`.
        :return: the namespace of (dense) for the output layer.
        """
        layer = SimpleNamespace(
            dense=gluon.nn.Dense(
                units=config.num_class,
                flatten=config.flatten
            )
        )

        with self.name_scope():
            self.__setattr__(layer.dense.name, layer.dense)

        return layer

    def _init_fuse_conv_layer(self, config: SimpleNamespace) -> Optional[SimpleNamespace]:
        """
        :param config: the output of :meth:`FFNNModel.namespace_input_conv_layer`.
        :param input_col: the column dimension of the input matrix.
        :return: the namespace of (conv, dropout) for the fuse convolution layer.
        """
        if config is None:
            return None
        layer = SimpleNamespace(
            conv=gluon.nn.Conv2D(
                channels=config.filters,
                kernel_size=(1, self.input_layer.col),
                strides=(1, self.input_layer.col),
                activation=config.activation),
            dropout=gluon.nn.Dropout(config.dropout))

        with self.name_scope():
            self.__setattr__(layer.conv.name, layer.conv)
            self.__setattr__(layer.dropout.name, layer.dropout)

        return layer

    def _init_ngram_conv_layers(self, config: SimpleNamespace) -> Optional[List[SimpleNamespace]]:

        if config is None:
            return None

        def conv(ngram):
            return gluon.nn.Conv2D(
                channels=config.filters,
                kernel_size=(ngram, self.input_layer.col if self.fuse_conv_layer is None else self.fuse_conv_layer._channels),
                strides=(1, self.input_layer.col),
                activation=config.activation)

        def pool(ngram):
            pool_model = gluon.nn.MaxPool2D if config.pool == 'max' else gluon.nn.AvgPool2D
            return pool_model(pool_size=(ngram, 1), strides=(ngram, 1))

        def layers(ngram):
            layer = SimpleNamespace(
                conv=conv(ngram),
                dropout=gluon.nn.Dropout(config.dropout),
                pool=pool(ngram) if config.pool is not None else None
            )

            with self.name_scope():
                self.__setattr__(layer.conv.name, layer.conv)
                self.__setattr__(layer.dropout.name, layer.dropout)
                if config.pool is not None:
                    self.__setattr__(layer.pool.name, layer.pool)

            return layer

        return [layers(ngram) for ngram in config.ngrams]

    def _init_hidden_layers(self, configs: Optional[Sequence[SimpleNamespace]]) -> Optional[List[SimpleNamespace]]:
        """
        :param configs: the sequence of outputs of :meth:`FFNNModel.namespace_hidden_layers`.
        :return: the namespace of (dense, dropout) for the hidden layers.
        """

        if configs is None:
            return None

        def hidden(c: SimpleNamespace) -> SimpleNamespace:
            layer = SimpleNamespace(
                dense=gluon.nn.Dense(units=c.dim,
                                     activation=c.activation),
                dropout=gluon.nn.Dropout(c.dropout))

            with self.name_scope():
                self.__setattr__('hidden_{}'.format(layer.dense.name), layer.dense)
                self.__setattr__('hidden_{}'.format(layer.dropout.name), layer.dropout)

            return layer

        return [hidden(config) for config in configs]


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import glob
import json
import logging
import os
from typing import List, Dict, Any, Tuple

from elit.dataset import LabelMap
from elit.structure import Sentence, TOK, Document, to_gold, SID, DOC_ID

__author__ = "Gary Lai"


def tsv_cols(tsv_heads: list) -> Dict[str, int]:
    d = {}
    for head in tsv_heads:
        d[head[0]] = head[1]
    return d


def tsv_reader(tsv_directory: str,
               cols: Dict[str, int],
               key: str = None) -> Tuple[List[Document], LabelMap]:
    documents = []
    wc = sc = 0
    label_map = LabelMap()

    if TOK not in cols:
        raise ValueError('The column index of "%s" must be specified' % TOK)

    if key is not None:
        if key in cols:
            cols = cols.copy()
            cols[to_gold(key)] = cols.pop(key)
        else:
            raise ValueError('Key mismatch: %s is not a key in %s' % (key, str(cols)))

    logging.info('Reading tsv from:')
    logging.info('- directory: %s' % tsv_directory)

    for filename in glob.glob('{}/*'.format(tsv_directory)):
        # avoid reading unexpected files, such as hidden files.
        if not os.path.isfile(filename):
            continue
        logging.info('  - file: %s' % filename)

        sentences = []
        sid = 0
        fields = {k: [] for k in cols.keys()}

        with open(filename) as fin:
            for line in fin.readlines():
                if line.startswith('#'):
                    continue
                l = line.split()

                if l:
                    for k, v in fields.items():
                        v.append(l[cols[k]])
                elif len(fields[TOK]) > 0:
                    wc += len(fields[TOK])
                    sentences.append(Sentence(fields))
                    fields = {k: [] for k in cols.keys()}

            if len(fields[TOK]) > 0:
                wc += len(fields[TOK])
                sentences.append(Sentence(fields))

        [[label_map.add(i) for i in sent[to_gold(key)]] for sent in sentences]
        [sent.update({SID: i}) for i, sent in enumerate(sentences)]
        sc += len(sentences)
        documents.append(Document(sens=sentences))

    [sent.update({DOC_ID: i}) for i, sent in enumerate(documents)]
    logging.info('- dc = %d, sc = %d, wc = %d' % (len(documents), sc, wc))
    return documents, label_map


def json_reader(filepath: str,
                cols: Any = None,
                key: str = None) -> Tuple[List[Document], LabelMap]:
    # TODO: update this to accept any format (see tsv_reader)
    documents = []
    dc = wc = sc = 0
    label_map = LabelMap()

    logging.info('Reading json file from: ')
    if not os.path.isfile(filepath) and filepath.endswith('.json'):
        raise ValueError("{} is not a valid format".format(filepath))
    logging.info('- filepath: %s' % filepath)

    with open(filepath) as f:
        docs = json.load(f)
        for i, doc in enumerate(docs):
            sentences = []
            for sen in doc['sens']:
                wc += len(sen['tok'])
                if key is not None:
                    sen = sen.copy()
                    sen[to_gold(key)] = sen.pop(key)
                sentences.append(Sentence(sen))
            sc += len(sentences)
            [[label_map.add(i) for i in sent[to_gold(key)]] for sent in sentences]
            document = Document(sens=sentences)
            document[DOC_ID] = i
            documents.append(document)
            dc += len(documents)
    logging.info('- dc = %d, sc = %d, wc = %d' % (dc, sc, wc))
    return documents, label_map



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc

import inspect

__author__ = "Gary Lai, Jinho D. Choi"

from elit.cli import BaseCLI


class ComponentCLI(BaseCLI):
    """
    :class:`ComponentCLI` is an abstract class to implement a command-line interface for a component.

    Abstract methods to be implemented:
      - :meth:`ComponentCLI.train`
      - :meth:`ComponentCLI.decode`
      - :meth:`ComponentCLI.evaluate`
    """

    def __init__(self, name: str, description: str = None):
        """
        :param name: the name of the component.
        :param description: the description of this component; if ``None``, the name is used instead.
        """
        usage = """elit {} <command> [<args>]
        
        commands:
            train: train a model (gold labels required)
            decode: predict labels (gold labels not required)
            evaluate: evaluate the pre-trained model (gold labels required)
        """.format(name)

        super().__init__(name=name, usage=usage, description=description)

    @classmethod
    @abc.abstractmethod
    def train(cls):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Trains a model for this component.
        """
        raise NotImplementedError('%s.%s()' % (cls.__class__.__name__, inspect.stack()[0][3]))

    @classmethod
    @abc.abstractmethod
    def decode(cls):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Predicts labels using this component.
        """
        raise NotImplementedError('%s.%s()' % (cls.__class__.__name__, inspect.stack()[0][3]))

    @classmethod
    @abc.abstractmethod
    def evaluate(cls):
        """
        :param args: the command-line arguments to be parsed by :class:`argparse.ArgumentParser`.

        Evaluates the current model of this component.
        """
        raise NotImplementedError('%s.%s()' % (cls.__class__.__name__, inspect.stack()[0][3]))



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from typing import Any

__author__ = 'Jinho D. Choi, Gary Lai'


class Component(abc.ABC):
    """
    :class:`Component` is an abstract class; any component developed in ELIT must inherit this class.

    Abstract methods to be implemented:
      - :meth:`Component.load`
      - :meth:`Component.save`
      - :meth:`Component.train`
      - :meth:`Component.decode`
      - :meth:`Component.evaluate`
    """

    @abc.abstractmethod
    def load(self, model_path: str, **kwargs):
        """
        :param model_path: the filepath where a model can be loaded.

        Loads a model for this component from the filepath.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def save(self, model_path: str, **kwargs):
        """
        :param model_path: the filepath where the current model can be saved.

        Saves the current model of this component to the filepath.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def train(self, trn_data: Any, dev_data: Any, model_path: str, **kwargs) -> float:
        """
        :param trn_data: training data.
        :param dev_data: development (validation) data.
        :param model_path: the filepath where the trained model(s) are to be saved.
        :return: the best score form the development data.

        Trains a model for this component and saves the model to the filepath.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def decode(self, data: Any, **kwargs):
        """
        :param data: input data.

        Processes the input data, make predictions, and saves the predicted labels back to the
        input data.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def evaluate(self, data: Any, **kwargs):
        """
        :param data: input data.

        Evaluates the current model of this component with the input data.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))





# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from typing import Sequence

from elit.component import Component
from elit.structure import Document

__author__ = "Gary Lai"


class NLPComponent(Component):
    """
    :class:`NLPComponent` is an abstract class; any NLP component developed in ELIT must inherit
    this class.
    It is similar to :class:`Component` except that the type of training and development data is
    specified to :class:`elit.structure.Document`.

    Abstract methods to be implemented:
      - :meth:`Component.init`
      - :meth:`Component.load`
      - :meth:`Component.save`
      - :meth:`NLPComponent.train`
      - :meth:`NLPComponent.decode`
      - :meth:`NLPComponent.evaluate`
    """

    @abc.abstractmethod
    def train(self, trn_docs: Sequence[Document], dev_docs: Sequence[Document], model_path: str, **kwargs) -> float:
        """
        :param trn_docs: the sequence of documents for training.
        :param dev_docs: the sequence of documents for development (validation).
        :param model_path: the filepath where trained model(s) are to be saved.
        :return: the best score form the development data.

        Trains a model for this component and saves the model to the filepath.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def decode(self, docs: Sequence[Document], **kwargs):
        """
        :param docs: the sequence of input documents.

        Processes the input documents, make predictions, and saves the predicted labels back to the
        input documents.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @abc.abstractmethod
    def evaluate(self, docs: Sequence[Document], **kwargs):
        """
        :param docs: the sequence of input documents.

        Evaluates the current model of this component with the input documents.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import Tuple

import abc
from mxnet import metric

from elit.structure import BILOU

__author__ = 'Jinho D. Choi'


class EvalMetric(metric.EvalMetric):
  
    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)

    @abc.abstractmethod
    def update(self, labels, preds):
        pass


class F1(EvalMetric):
  
    def __init__(self):
        super().__init__(name='f1score')
        self.correct = 0.0
        self.p_total = 0.0
        self.r_total = 0.0

    def precision(self) -> float:
        return self.correct / self.p_total if self.p_total > 0.0 else 0.0

    def recall(self) -> float:
        return self.correct / self.r_total if self.r_total > 0.0 else 0.0

    def f1(self) -> float:
        p = self.precision()
        r = self.recall()
        return 2 * p * r / (p + r) if p + r > 0.0 else 0.0

    def reset(self):
        self.correct = 0.0
        self.p_total = 0.0
        self.r_total = 0.0

    def get(self) -> Tuple[str, float]:
        return self.name, self.f1()

    @abc.abstractmethod
    def update(self, labels, preds):
        pass


class ChunkF1(F1):

    def update(self, labels, preds):
        gold = BILOU.to_chunks(labels)
        pred = BILOU.to_chunks(preds)
        self.correct += len(set.intersection(set(gold), set(pred)))
        self.p_total += len(pred)
        self.r_total += len(gold)



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import logging
import sys

__author__ = "Gary Lai"


def set_logger(filename: str = None, level: int = logging.INFO, formatter: logging.Formatter = None):
    log = logging.getLogger()
    log.setLevel(level)
    ch = logging.StreamHandler(sys.stdout) if filename is None else logging.FileHandler(filename)
    if formatter is not None:
        ch.setFormatter(formatter)
    log.addHandler(ch)



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import configparser
import os

__author__ = "Han He"


class Config(object):
    def __init__(self, config_file: str, extra_args=None):
        if not os.path.isfile(config_file):
            raise FileNotFoundError('config file {} not found'.format(config_file))
        _config = configparser.ConfigParser(allow_no_value=True)
        _config.read(config_file)
        if extra_args:
            for section in _config.sections():
                for k, v in _config.items(section):
                    if k in extra_args:
                        v = type(v)(extra_args[k])
                        _config.set(section, k, v)
        self._config = _config

    @property
    def config(self):
        return self._config


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import argparse
import sys

__author__ = "Gary Lai"


class ElitCli(object):
    def __init__(self):
        # set_logger()
        usage = '''elit <command> [<args>]

        commands:
            token_tagger: use token tagger
            install: Install model
            download: download pre-trained model or word embedding
        '''
        parser = argparse.ArgumentParser(usage=usage)
        parser.add_argument('command', help='command to run')
        args = parser.parse_args(sys.argv[1:2])
        if args.command == 'token_tagger':
            from elit.nlp.token_tagger import TokenTaggerCLI
            TokenTaggerCLI()
        elif args.command == 'install':
            from elit.cli.install import InstallCLI
            InstallCLI()
        elif args.command == 'download':
            from elit.cli.download import DownloadCLI
            DownloadCLI()
        else:
            print('Unrecognized command')
            parser.print_help()
            exit(1)


from .base import BaseCLI
from .component import ComponentCLI




# THIS FILE IS GENERATED FROM SCIPY SETUP.PY
short_version = '0.1.29'
version = '0.1.29'
full_version = '0.1.29.dev1550246703'
dev_version = '1550246703'
release = False
if not release:
    version = full_version



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
import os
import re
from typing import List, Tuple

from pkg_resources import resource_filename

from elit.component import Component
from elit.structure import TOK, OFF, Document, Sentence, SID
from elit.util.io import read_word_set, read_concat_word_dict
from elit.util.string import *

__author__ = "Jinho D. Choi, Gary Lai"


class Tokenizer(Component):
    def decode(self, input_text: str, init_offset: int = 0, segment: int = 2, **kwargs) -> Document:
        """
        :param input_text: the input text.
        :param init_offset: the initial offset of the first token.
        :param segment: 0 - no segmentation, 1 - newline segmentation, 2 - rule-based segmentation, 3 - newline and rule-based segmentation.
        :return: the dictionary contains ('tok' = list of tokens) and ('off' = list of offsets);
                 see the comments for :meth:`Tokenizer.offsets` for more details about the offsets.
        """
        document = Document()

        if segment == 0:
            tokens, offsets = self.tokenize(input_text, init_offset)
            document.add_sentence(Sentence({TOK: tokens, OFF: offsets}))
        elif segment == 2:
            tokens, offsets = self.tokenize(input_text, init_offset)
            document.add_sentences(self.segment(tokens, offsets))
        elif segment == 1 or segment == 3:
            indices = [i for i, c in enumerate(input_text) if i == 0 or (c == '\n' and input_text[i - 1] != '\n')]
            indices.append(len(input_text))

            for i in range(1, len(indices)):
                bidx = indices[i - 1]
                eidx = indices[i]
                tokens, offsets = self.tokenize(input_text[bidx:eidx], bidx+init_offset)
                if tokens:
                    if segment == 1: document.add_sentence(Sentence({TOK: tokens, OFF: offsets}))
                    else: document.add_sentences(self.segment(tokens, offsets))

        for i, sentence in enumerate(document.sentences): sentence[SID] = i
        return document

    @abc.abstractmethod
    def tokenize(self, input_text: str, init_offset: int = 0) -> Tuple[List[str], List[Tuple[int, int]]]:
        """
        :param input_text: the input text.
        :param init_offset: the initial offset of the first token.
        :return: the dictionary contains ('tok' = list of tokens) and ('off' = list of offsets);
                 see the comments for :meth:`Tokenizer.offsets` for more details about the offsets.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    @classmethod
    def get_offsets(cls, input_text: str, tokens: List[str], init_offset=0) -> List[Tuple[int, int]]:
        """
        :param input_text: the input text.
        :param tokens: the list of tokens split from the input text.
        :param init_offset: the initial offset of the first token.
        :return: the list of (begin, end) offsets, where the begin (inclusive) and the end (exclusive) offsets indicate
        the caret positions of the first and the last characters of the corresponding token, respectively.
        e.g., text = 'Hello, world!', tokens = ['Hello', ',', 'world', '!'] -> [(0, 5), (5, 6), (7, 12), (12, 13)]
        """

        def get_offset(token):
            nonlocal end
            begin = input_text.index(token, end)
            end = begin + len(token)
            return begin + init_offset, end + init_offset

        end = 0
        return [get_offset(token) for token in tokens]

    @classmethod
    def segment(cls, tokens: List[str], offsets: List[Tuple[int, int]]) -> List[Sentence]:
        """
        :param tokens: the list of input tokens.
        :param offsets: the list of offsets, where each offset is a tuple of (begin, end).
        :return: the list of sentences segmented from the input text.
        """

        def sentence(bidx: int, eidx: int) -> Sentence:
            return Sentence({TOK: tokens[bidx:eidx], OFF: offsets[bidx:eidx]})

        right_quote = True
        sentences = []
        begin = 0

        for i, token in enumerate(tokens):
            t = token[0]
            if t == '"': right_quote = not right_quote

            if begin == i:
                if sentences and (is_right_bracket(t) or t == u'\u201D' or t == '"' and right_quote):
                    d = sentences[-1]
                    d[TOK].append(token)
                    d[OFF].append(offsets[i])
                    begin = i + 1
            elif all(is_final_mark(c) for c in token):
                sentences.append(sentence(begin, i + 1))
                begin = i + 1

        if begin < len(tokens):
            sentences.append(sentence(begin, len(tokens)))

        return sentences


class SpaceTokenizer(Tokenizer):
    """
    :class:`SpaceTokenizer` splits tokens by white-spaces.
    """

    def __init__(self):
        super(SpaceTokenizer, self).__init__()

    def save(self, model_path: str, **kwargs):
        """ Not supported. """
        pass

    def decode(self, input_text: str, init_offset: int = 0, segment: int = 1, **kwargs) -> Document:
        return super().decode(input_text, init_offset, segment, **kwargs)

    def train(self, trn_data, dev_data, model_path: str, **kwargs) -> float:
        """ Not supported. """
        pass

    def evaluate(self, data, **kwargs):
        """ Not supported. """
        pass

    def save(self, model_path: str, **kwargs):
        """ Not supported. """
        pass

    def decode(self, input_text: str, init_offset: int = 0, segment: int = 1, **kwargs) -> Document:
        return super().decode(input_text, init_offset, segment, **kwargs)

    def train(self, trn_data, dev_data, model_path: str, **kwargs) -> float:
        """ Not supported. """
        pass

    def evaluate(self, data, **kwargs):
        """ Not supported. """
        pass

    def load(self, model_path: str, **kwargs):
        """ Not supported. """
        pass

    def tokenize(self, input_text: str, init_offset: int = 0) -> Tuple[List[str], List[Tuple[int, int]]]:
        """
        :param input_text: the input text.
        :param init_offset: the initial offset of the first token.
        :return: the dictionary contains ('tok' = list of tokens) and ('off' = list of offsets);
                 see the comments for :meth:`Tokenizer.offsets` for more details about the offsets.
        """
        tokens = input_text.split()
        return tokens, self.get_offsets(input_text, tokens, init_offset)


class EnglishTokenizer(Tokenizer):
    def __init__(self):
        """
        :class:`EnglishTokenizer` splits the input text into linguistic tokens.
        """
        super(EnglishTokenizer, self).__init__()

        # _inflection_lexicons
        resource_root = 'elit.resources'

        self.ABBREVIATION_PERIOD = read_word_set(resource_filename(
            resource_root, 'english_abbreviation_period.txt'))
        self.APOSTROPHE_FRONT = read_word_set(resource_filename(
            resource_root, 'english_apostrophe_front.txt'))
        self.MAP_CONCAT_WORD = read_concat_word_dict(resource_filename(
            resource_root, 'english_concat_words.txt'))
        self.HYPHEN_PREFIX = read_word_set(resource_filename(
            resource_root, 'english_hyphen_prefix.txt'))
        self.HYPHEN_SUFFIX = read_word_set(resource_filename(
            resource_root, 'english_hyphen_suffix.txt'))

        # regular expressions
        self.RE_NETWORK_PROTOCOL = re.compile(
            r'((http|https|ftp|sftp|ssh|ssl|telnet|smtp|pop3|imap|imap4|sip)(://))')
        """
        :abc:
        <3 </3 <\3
        (: ): \\: *: $: (-: (^: (= (;
        :) :( =) B) 8) :-) :^) :3 :D :p :| :(( :---)
        """
        self.RE_EMOTICON = re.compile(
            r'(:\w+:|<[\\/]?3|[()\\|*$][-^]?[:=;]|[:=;B8]([-^]+)?[3DOPp@$*()\\/|]+)(\W|$)')
        """
        jinho@elit.cloud
        jinho.choi@elit.cloud
        choi@demo.elit.cloud
        jinho:choi@127.0.0.1
        """
        self.RE_EMAIL = re.compile(
            r'[\w\-.]+(:\S+)?@(([A-Za-z0-9\-]+\.)+[A-Za-z]{2,12}|\d{1,3}(\.\d{1,3}){3})')
        """
        &arrow;
        &#123; &#x123; &#X123;
        """
        self.RE_HTML_ENTITY = re.compile(r'&([A-Za-z]+|#[Xx]?\d+);')
        """
        [1] (1a) {A} <a1> [***] [A.a] [A.1] [1.a] ((---))
        """
        self.RE_LIST_ITEM = re.compile(
            r'(([\[({<]+)(\d+[A-Za-z]?|[A-Za-z]\d*|\W+)(\.(\d+|[A-Za-z]))*([\])\}>])+)')
        """
        don't don’t I'll HE'S
        """
        self.RE_APOSTROPHE = re.compile(
            r'(?i)[a-z](n[\'\u2019]t|[\'\u2019](ll|nt|re|ve|[dmstz]))(\W|$)')
        """
        a.b.c 1-2-3
        """
        self.RE_ABBREVIATION = re.compile(r'[A-Za-z0-9]([.-][A-Za-z0-9])*$')
        """
        10kg 1cm
        """
        self.RE_UNIT = re.compile(
            r'(?i)(\d)([acdfkmnpyz]?[mg]|[ap]\.m|ch|cwt|d|drc|ft|fur|gr|h|in|lb|lea|mi|ms|oz|pg|qtr|yd)$')
        """
        hello.World
        """
        self.RE_FINAL_MARK_IN_BETWEEN = re.compile(
            r'([A-Za-z]{3,})([.?!]+)([A-Za-z]{3,})$')

    def save(self, model_path: str, **kwargs):
        """ Not supported. """
        pass

    def train(self, trn_data, dev_data, model_path: str, **kwargs) -> float:
        """ Not supported. """
        pass

    def evaluate(self, data, **kwargs):
        """ Not supported. """
        pass

    def load(self, model_path, *args, **kwargs):
        self.ABBREVIATION_PERIOD = read_word_set(
            os.path.join(model_path, 'english_abbreviation_period.txt'))
        self.APOSTROPHE_FRONT = read_word_set(
            os.path.join(model_path, 'english_apostrophe_front.txt'))
        self.MAP_CONCAT_WORD = read_concat_word_dict(
            os.path.join(model_path, 'english_concat_words.txt'))
        self.HYPHEN_PREFIX = read_word_set(
            os.path.join(model_path, 'english_hyphen_prefix.txt'))
        self.HYPHEN_SUFFIX = read_word_set(
            os.path.join(model_path, 'english_hyphen_suffix.txt'))

    def tokenize(self, input_text: str, init_offset: int = 0) -> Tuple[List[str], List[Tuple[int, int]]]:
        """
        :param input_text: the input text.
        :param init_offset: the initial offset of the first token.
        :return: the dictionary contains ('tok' = list of tokens) and ('off' = list of offsets);
                 see the comments for :meth:`Tokenizer.offsets` for more details about the offsets.
        """
        tokens = []
        offsets = []

        # no valid token in the input text
        if not input_text or input_text.isspace():
            return tokens, offsets

        # skip beginning and ending spaces
        begin = next(i for i, c in enumerate(input_text) if not c.isspace())
        last = len(input_text) - next(i for i, c in enumerate(reversed(input_text)) if not c.isspace())

        # search for in-between spaces
        for end, c in enumerate(input_text[begin + 1:last], begin + 1):
            if c.isspace():
                self.tokenize_aux(tokens, offsets, input_text, begin, end, init_offset)
                begin = end + 1

        self.tokenize_aux(tokens, offsets, input_text, begin, last, init_offset)
        return tokens, offsets

    def tokenize_aux(self, tokens, offsets, text, begin, end, offset):
        if begin >= end or end > len(text): return False
        token = text[begin:end]

        # handle special cases
        if self.tokenize_trivial(tokens, offsets, token, begin, end, offset):
            return True
        if self.tokenize_regex(tokens, offsets, text, begin, end, offset, token):
            return True
        if self.tokenize_symbol(tokens, offsets, text, begin, end, offset, token):
            return True

        # add the token as it is
        self.add_token(tokens, offsets, token, begin, end, offset)
        return True

    def tokenize_trivial(self, tokens, offsets, token, begin, end, offset):
        if end - begin == 1 or token.isalnum():
            self.add_token(tokens, offsets, token, begin, end, offset)
            return True

        return False

    def tokenize_regex(self, tokens, offsets, text, begin, end, offset, token):
        def group(regex, gid=0):
            m = regex.search(token)
            if m:
                idx = begin + m.start(gid)
                lst = begin + m.end(gid)

                self.tokenize_aux(tokens, offsets, text, begin, idx, offset)
                self.add_token(tokens, offsets, m.group(gid), idx, lst, offset)
                self.tokenize_aux(tokens, offsets, text, lst, end, offset)
                return True

            return False

        def hyperlink():
            m = self.RE_NETWORK_PROTOCOL.search(token)
            if m:
                if m.start() > 0:
                    idx = begin + m.start()
                    self.tokenize_aux(tokens, offsets, text, begin, idx, offset)
                    self.add_token(tokens, offsets, token[m.start():], idx, end, offset)
                else:
                    self.add_token(tokens, offsets, token, begin, end, offset)
                return True

            return False

        # split by regular expressions
        if group(self.RE_HTML_ENTITY):
            return True
        if group(self.RE_EMAIL):
            return True
        if hyperlink():
            return True
        if group(self.RE_EMOTICON, 1):
            return True
        if group(self.RE_LIST_ITEM):
            return True
        if group(self.RE_APOSTROPHE, 1):
            return True
        return False

    def tokenize_symbol(self, tokens, offsets, text, begin, end, offset, token):
        def index_last_sequence(i, c):
            final_mark = is_final_mark(c)

            for j, d in enumerate(token[i + 1:], i + 1):
                if final_mark:
                    if not is_final_mark(d):
                        return j
                elif c != d:
                    return j

            return len(token)

        def skip(i, c):
            if c == '.' or c == '+':  # .1, +1
                return self.is_digit(token, i + 1)
            if c == '-':  # -1
                return i == 0 and self.is_digit(token, i + 1)
            if c == ',':  # 1,000,000
                return self.is_digit(token, i - 1) and self.is_digit(token, i + 1, i + 4) and not self.is_digit(token, i + 4)
            if c == ':':
                # 1:2
                return self.is_digit(token, i - 1) and self.is_digit(token, i + 1)
            if is_single_quote(c):
                return self.is_digit(token, i + 1, i + 3) and not self.is_digit(token, i + 3)  # '97
            return False

        def split(i, c, p0, p1):
            if p0(c):
                j = index_last_sequence(i, c)

                if p1(i, j):
                    idx = begin + i
                    lst = begin + j

                    self.tokenize_aux(tokens, offsets, text, begin, idx, offset)
                    self.add_token(tokens, offsets, token[i:j], idx, lst, offset)
                    self.tokenize_aux(tokens, offsets, text, lst, end, offset)
                    return True

            return False

        def separator_0(c):
            return c in {',', ';', ':', '~', '&', '|', '/'} or is_bracket(c) or is_arrow(c) or is_double_quote(c) or is_hyphen(c)

        def edge_symbol_0(c):
            return is_single_quote(c) or is_final_mark(c)

        def currency_like_0(c):
            return c == '#' or is_currency(c)

        def edge_symbol_1(i, j):
            return i + 1 < j or i == 0 or j == len(token) or is_punct(token[i - 1]) or is_punct(token[j])

        def currency_like_1(i, j):
            return i + 1 < j or j == len(token) or token[j].isdigit()

        # split by symbols
        for i, c in enumerate(token):
            if skip(i, c):
                continue
            if split(i, c, separator_0, lambda i, j: True):
                return True
            if split(i, c, edge_symbol_0, edge_symbol_1):
                return True
            if split(i, c, currency_like_0, currency_like_1):
                return True
        return False

    def add_token(self, tokens, offsets, token, begin, end, offset):
        if not self.concat_token(tokens, offsets, token, end) and not self.split_token(tokens, offsets, token, begin, end, offset):
            self.add_token_aux(tokens, offsets, token, begin, end, offset)

    def concat_token(self, tokens, offsets, token, end):
        def apostrophe_front(prev, curr):
            return len(prev) == 1 and is_single_quote(prev) and curr in self.APOSTROPHE_FRONT

        def abbreviation(prev, curr):
            return curr == '.' and (self.RE_ABBREVIATION.match(prev) or prev in self.ABBREVIATION_PERIOD)

        def acronym(prev, curr, next):
            return len(curr) == 1 and curr in {'&', '|', '/'} and (len(prev) <= 2 and len(next) <= 2 or prev.isupper() and next.isupper())

        def hyphenated(prev, curr, next):
            p = len(prev)

            if len(curr) == 1 and is_hyphen(curr):
                if self.is_digit(prev, p - 3, p) and (p == 3 or is_hyphen(prev[p - 4])) and next.isdigit():
                    # 000-0000, 000-000-0000
                    return True
                if prev[-1].isalnum() and (len(prev) == 1 or is_hyphen(prev[p - 2])) and len(next) == 1 and next.isalnum():
                    # p-u-s-h
                    return True
                return (prev in self.HYPHEN_PREFIX and next.isalnum()) or (next in self.HYPHEN_SUFFIX and prev.isalnum())

            return False

        def coloned(prev, curr, next):
            if prev == 're' and curr == ':' and next == 'invent':
                # re:Invent
                return True

            return False

        def no_dot_digit(prev, curr, next):
            if prev == 'no' and curr == '.' and next[0].isdigit():
                t, o = tokens.pop(), offsets.pop()
                tokens[-1] += t
                offsets[-1] = (offsets[-1][0], o[1])
                return True

            return False

        # concatenate split tokens if necessary
        if tokens:
            prev = tokens[-1].lower()
            curr = token.lower()

            if apostrophe_front(prev, curr) or abbreviation(prev, curr):
                tokens[-1] += token
                offsets[-1] = (offsets[-1][0], end)
                return True

        if len(tokens) >= 2:
            prev = tokens[-2].lower()
            curr = tokens[-1].lower()
            next = token.lower()

            if acronym(tokens[-2], curr, token) or hyphenated(prev, curr, next) or coloned(prev, curr, next):
                tokens[-2] += tokens[-1] + token
                offsets[-2] = (offsets[-2][0], end)
                del tokens[-1]
                del offsets[-1]
                return True

            no_dot_digit(prev, curr, next)

        return False

    def split_token(self, tokens, offsets, token, begin, end, offset):
        def unit():
            m = self.RE_UNIT.search(token)
            if m:
                idx = begin + m.start(2)
                self.add_token_aux(tokens, offsets, token[:m.start(2)], begin, idx, offset)
                self.add_token_aux(tokens, offsets, m.group(2), idx, end, offset)
                return True
            return False

        def concat_words():
            t = self.MAP_CONCAT_WORD.get(token.lower(), None)
            if t:
                i = 0
                for j in t:
                    self.add_token_aux(tokens, offsets, token[i:j], begin + i, begin + j, offset)
                    i = j
                return True
            return False

        def final_mark():
            m = self.RE_FINAL_MARK_IN_BETWEEN.match(token)
            if m:
                for i in range(1, 4):
                    self.add_token_aux(tokens, offsets, m.group(i), begin + m.start(i), begin + m.end(i), offset)
                return True
            return False

        return unit() or concat_words() or final_mark()

    @staticmethod
    def add_token_aux(tokens, offsets, token, begin, end, offset):
        tokens.append(token)
        offsets.append((begin + offset, end + offset))

    @staticmethod
    def is_digit(token, i, j=None):
        if 0 <= i < len(token):
            if j is None:
                return token[i].isdigit()
            if i < j <= len(token):
                return token[i:j].isdigit()
        return False


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import logging

import numpy as np
from gensim.models import KeyedVectors

from elit.embedding import TokenEmbedding

__author__ = "Gary Lai"


class Word2Vec(TokenEmbedding):
    """
    :class:`Word2Vec` is a token-based model trained by `Word2Vec <https://github.com/tmikolov/word2vec>`_.
    """

    def __init__(self, filepath: str):
        """
        :param filepath: the path to the binary file containing a word embedding model trained by Word2Vec (``*.bin``).
        """
        logging.info('Word2Vec')
        logging.info('- model: {}'.format(filepath))
        self.model = KeyedVectors.load(filepath) if filepath.lower().endswith('.gnsm') else KeyedVectors.load_word2vec_format(filepath, binary=True)
        dim = self.model.syn0.shape[1]
        super().__init__(dim)
        logging.info('- vocab = %d, dim = %d' % (len(self.model.vocab), dim))

    # override
    def emb(self, value: str) -> np.ndarray:
        vocab = self.model.vocab.get(value, None)
        return self.model.syn0[vocab.index] if vocab else self.pad


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from typing import List, Union, Sequence, Tuple

import abc
import numpy as np
from gluonnlp.data.batchify import batchify
from mxnet import nd, gluon
from mxnet.ndarray import NDArray
from tqdm import tqdm

from elit.embedding import Embedding
from elit.structure import to_gold, Sentence, Document

__author__ = "Gary Lai"


class LabelMap(object):
    """
    :class:`LabelMap` provides the mapping between string labels and their unique integer IDs.
    """

    def __init__(self):
        self.index_map = {}
        self.labels = {}

    def __len__(self):
        return len(self.labels)

    def __str__(self):
        return str(self.index_map)

    def add(self, label: str) -> int:
        """
        :param label: the label.
        :return: the class ID of the label.

        Adds the label to this map and assigns its class ID if not already exists.
        """
        idx = self.cid(label)
        if idx < 0:
            idx = len(self.labels)
            self.index_map[label] = idx
            self.labels[idx] = label
        return idx

    def get(self, cid: int) -> str:
        """
        :param cid: the class ID.
        :return: the label corresponding to the class ID.
        """
        return self.labels.get(cid, None)

    def cid(self, label: str) -> int:
        """
        :param label: the label.
        :return: the class ID of the label if exists; otherwise, -1.
        """
        return self.index_map.get(label, -1)

    def num_class(self) -> int:
        return len(self.labels)

    def argmax(self, scores: Union[np.ndarray, NDArray]) -> str:
        """
        :param scores: the prediction scores of all labels.
        :return: the label with the maximum score.
        """
        if self.__len__() < len(scores):
            scores = scores[:self.__len__()]
        n = nd.argmax(scores, axis=0).asscalar() if isinstance(scores, NDArray) else np.argmax(scores)
        return self.get(int(n))


class Dataset(gluon.data.Dataset):

    def __init__(self, docs: Sequence[Document], embs: List[Union[Embedding]], key: str, label_map: LabelMap, label: bool = True, transform=None):
        self._data = []
        self.embs = embs
        self.key = key
        self.label_map = label_map
        self.label = label
        self.transform = transform
        self.docs = docs
        self.init_data()

    def __getitem__(self, idx):
        if self.transform is not None:
            return self.transform(self._data[idx])
        else:
            return self._data[idx]

    def __len__(self):
        return len(self._data)

    @abc.abstractmethod
    def extract(self, did: int, sid: int, sen: Sentence):
        raise NotImplementedError

    def extract_sen(self, sen):
        return nd.array([np.concatenate(i) for i in zip(*[emb.emb_list(sen.tokens) for emb in self.embs])]).reshape(0, -1)

    def init_data(self):
        for did, doc in enumerate(tqdm(self.docs, leave=False)):
            for sid, sen in enumerate(tqdm(doc.sentences, leave=False)):
                self.extract(did, sid, sen)


class TokensDataset(Dataset):

    def __init__(self, docs: Sequence[Document], embs: List[Embedding], key: str, label_map: LabelMap, feature_windows: Tuple, label: bool = True, transform=None):
        self.embs = embs
        self.feature_windows = feature_windows
        self.pad = nd.zeros(sum([emb.dim for emb in self.embs]))
        super().__init__(docs=docs, embs=embs, key=key, label_map=label_map, label=label, transform=transform)

    def extract_x(self, i, w):
        return nd.stack(*[w[i + win] if 0 <= (i + win) < len(w) else self.pad for win in self.feature_windows])

    def extract_y(self, label: str):
        return self.label_map.cid(label)

    def extract(self, did, sid, sen, **kwargs):
        w = self.extract_sen(sen)
        if self.label:
            for i, label in enumerate(sen[to_gold(self.key)]):
                self._data.append((self.extract_x(i, w), self.extract_y(label)))
        else:
            for i in range(len(sen)):
                self._data.append((self.extract_x(i, w), -1))


class SequencesDataset(Dataset):

    def __init__(self, docs: Sequence[Document], embs: List[Embedding], key: str, label_map: LabelMap, label: bool = True, transform=None):
        super().__init__(docs=docs, embs=embs, key=key, label_map=label_map, label=label, transform=transform)

    def extract_labels(self, sen):
        if self.label:
            return nd.array([self.label_map.cid(l) for l in sen[to_gold(self.key)]])
        else:
            return nd.array([-1 for _ in range(len(sen))])

    def extract(self, did, sid, sen):
        self._data.append((did, sid, self.extract_sen(sen), self.extract_labels(sen)))


sequence_batchify_fn = batchify.Tuple((batchify.Stack(), batchify.Stack(), batchify.Pad(), batchify.Pad(pad_val=-1)))


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

import mxnet as mx

__author__ = "Gary Lai"


def mx_loss(s: str) -> mx.gluon.loss:
    s = s.lower()

    if s == 'softmaxcrossentropyloss':
        return mx.gluon.loss.SoftmaxCrossEntropyLoss()
    if s == 'sigmoidbinarycrossentropyloss':
        return mx.gluon.loss.SigmoidBinaryCrossEntropyLoss()
    if s == 'l2loss':
        return mx.gluon.loss.L2Loss()
    if s == 'l2loss':
        return mx.gluon.loss.L1Loss()
    if s == 'kldivloss':
        return mx.gluon.loss.KLDivLoss()
    if s == 'huberloss':
        return mx.gluon.loss.HuberLoss()
    if s == 'hingeloss':
        return mx.gluon.loss.HingeLoss()
    if s == 'squaredhingeloss':
        return mx.gluon.loss.SquaredHingeLoss()
    if s == 'logisticloss':
        return mx.gluon.loss.LogisticLoss()
    if s == 'tripletloss':
        return mx.gluon.loss.TripletLoss()
    if s == 'ctcloss':
        return mx.gluon.loss.CTCLoss()

    raise TypeError("Unsupported loss: " + s)




# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import inspect
from typing import List, Sequence

import numpy as np

from elit.embedding import Embedding
from elit.structure import Document

__author__ = "Gary Lai"


class TokenEmbedding(Embedding):
    """
    :class:`TokenEmbedding` is an abstract class to implement token-based embedding models.

    Subclasses:
      - :class:`Word2Vec`
      - :class:`FastText`

    Abstract methods to be implemented:
      - :meth:`TokenEmbedding.emb`
    """

    def __init__(self, dim: int):
        """
        :param dim: the dimension of each embedding.
        """
        super().__init__(dim)

    # override
    def embed(self, docs: Sequence[Document], key: str, **kwargs):
        """
        Adds a list of embeddings to each sentence corresponding to its tokens.

        :param docs: a sequence of input documents.
        :param key: the key to each sentence where the list of embeddings is to be added.
        """
        for doc in docs:
            for sen in doc:
                sen[key] = self.emb_list(sen.tokens)

    @abc.abstractmethod
    def emb(self, token: str) -> np.ndarray:
        """
        :param token: the input token.
        :return: the embedding of the input token.
        """
        raise NotImplementedError('%s.%s()' % (self.__class__.__name__, inspect.stack()[0][3]))

    def emb_list(self, tokens: Sequence[str]) -> List[np.ndarray]:
        """
        :param tokens: the sequence of input tokens.
        :return: the list of embeddings for the corresponding tokens.
        """
        return [self.emb(value) for value in tokens]

    def emb_matrix(self, tokens: Sequence[str], maxlen: int) -> List[np.ndarray]:
        """
        :param tokens: the sequence of input tokens.
        :param maxlen: the maximum length of the output list;
                       if ``> len(values)``, the bottom part of the matrix is padded with zero embeddings;
                       if ``< len(values)``, embeddings of the exceeding values are discarded from the resulting matrix.
        :return: the matrix where each row is the embedding of the corresponding value.
        """
        return [self.emb(tokens[i]) if i < len(tokens) else self.pad for i in range(maxlen)]


# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import abc
import datetime
import logging
import time
from typing import Sequence, List

import mxnet as mx
from mxnet.gluon import Trainer
from mxnet.gluon.data import DataLoader
from tqdm import trange

from elit.component import NLPComponent
from elit.embedding import Embedding
from elit.structure import Document
from elit.util.mx import mx_loss

__author__ = "Gary Lai"


class MXComponent(NLPComponent):
    """
    :class:`TokenTagger` provides an abstract class to implement a tagger that predicts a tag for every token.
    """

    def __init__(self, ctx: mx.Context, key: str, embs: List[Embedding], **kwargs):
        """
        :param ctx: the (list of) device context(s) for :class:`mxnet.gluon.Block`.
        :param vsms: the sequence of namespace(model, key),
                         where the key indicates the key of the values to retrieve embeddings for (e.g., tok).
        """
        self.ctx = ctx
        self.key = key
        self.embs = embs
        self.model = None
        self.loss = None
        self.trainer = None

    def train(self, trn_docs: Sequence[Document], dev_docs: Sequence[Document], model_path: str,
              epoch: int = 100,
              trn_batch: int = 64,
              dev_batch: int = 2048,
              loss=mx_loss('softmaxcrossentropyloss'),
              optimizer='adagrad',
              optimizer_params=None,
              **kwargs) -> float:
        log = ('Training configuration',
               '- max epoch: {}'.format(epoch),
               '- trn batch: {}'.format(trn_batch),
               '- dev batch: {}'.format(dev_batch),
               '- loss func: {}'.format(loss),
               '- optimizer: {} <- {}'.format(optimizer, optimizer_params))
        logging.info('\n'.join(log))

        self.model.initialize(ctx=self.ctx, force_reinit=True)
        self.loss = loss
        self.trainer = Trainer(self.model.collect_params(), optimizer=optimizer, optimizer_params=optimizer_params)

        trn_data = self.data_loader(docs=trn_docs, batch_size=trn_batch, shuffle=True, label=True, bucket=True)
        dev_data = self.data_loader(docs=dev_docs, batch_size=dev_batch, shuffle=False, label=True, bucket=True)

        logging.info('Training')
        best_e, best_eval = -1, -1
        epochs = trange(1, epoch + 1)

        for e in epochs:
            trn_st = time.time()
            trn_acc = 100 * self.train_block(data_iter=trn_data, docs=trn_docs)
            trn_et = time.time()

            dev_st = time.time()
            dev_acc = 100 * self.evaluate_block(data_iter=dev_data, docs=dev_docs)
            dev_et = time.time()

            if best_eval < dev_acc:
                best_e, best_eval = e, dev_acc
                self.save(model_path=model_path)

            desc = ("epoch: {}".format(e),
                    "trn time: {}".format(datetime.timedelta(seconds=(trn_et - trn_st))),
                    "trn acc: {:.2f}".format(trn_acc),
                    "dev time: {}".format(datetime.timedelta(seconds=(dev_et - dev_st))),
                    "dev acc: {:.2f}".format(dev_acc),
                    "best epoch: {}".format(best_e),
                    "best eval: {:.2f}".format(best_eval))
            epochs.set_description(desc=' '.join(desc))
        return best_eval

    @abc.abstractmethod
    def train_block(self, data_iter: DataLoader, docs: Sequence[Document]) -> float:
        raise NotImplementedError

    def decode(self, docs: Sequence[Document], batch_size: int = 2048, **kwargs) -> Sequence[Document]:
        data_iter = self.data_loader(docs=docs, batch_size=batch_size, shuffle=False, label=False, bucket=True, **kwargs)
        self.decode_block(data_iter=data_iter, docs=docs)
        return docs

    @abc.abstractmethod
    def decode_block(self, data_iter: DataLoader, docs: Sequence[Document], **kwargs):
        raise NotImplementedError

    def evaluate(self, docs: Sequence[Document], batch_size: int = 2048, **kwargs) -> tuple:
        st = time.time()
        data_iter = self.data_loader(docs=docs, batch_size=batch_size, shuffle=False, label=True, bucket=True, **kwargs)
        acc = self.evaluate_block(data_iter=data_iter, docs=docs)
        et = time.time()
        return acc, et - st

    @abc.abstractmethod
    def evaluate_block(self, data_iter: DataLoader, docs: Sequence[Document]):
        raise NotImplementedError

    @abc.abstractmethod
    def data_loader(self, docs: Sequence[Document], batch_size, shuffle, **kwargs):
        raise NotImplementedError



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
import logging
from typing import Sequence, List

import mxnet as mx
from mxnet.gluon.data import DataLoader

from elit.component import MXComponent
from elit.embedding import Embedding
from elit.model import RNNModel
from elit.structure import Document

__author__ = "Gary Lai"


class RNNComponent(MXComponent):

    def __init__(self, ctx: mx.Context, key: str, embs: List[Embedding], rnn_config=None, output_config=None, **kwargs):
        super().__init__(ctx, key, embs, **kwargs)
        self.rnn_config = rnn_config
        self.output_config = output_config
        if rnn_config is not None and output_config is not None:
            self.model = RNNModel(rnn_config=rnn_config, output_config=output_config)
        else:
            self.model = None
        logging.info(self.__str__())

    def train_block(self, data_iter: DataLoader, docs: Sequence[Document]) -> float:
        pass

    def decode_block(self, data_iter: DataLoader, docs: Sequence[Document], **kwargs):
        pass

    def evaluate_block(self, data_iter: DataLoader, docs: Sequence[Document]):
        pass

    def data_loader(self, docs: Sequence[Document], batch_size, shuffle, **kwargs):
        pass

    def load(self, model_path: str, **kwargs):
        pass

    def save(self, model_path: str, **kwargs):
        pass



# ========================================================================
# Copyright 2018 ELIT
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================
from mxnet import gluon
from mxnet.gluon import Block
from types import SimpleNamespace

__author__ = "Gary Lai"


class RNNModel(Block):

    def __init__(self, rnn_config: SimpleNamespace, output_config: SimpleNamespace):
        super().__init__()
        self.rnn_layer = self._init_rnn_layer(rnn_config)
        self.output_layer = self._init_output_layer(output_config)

    def forward(self, X, state, *args):
        Y, state = self.rnn_layer.rnn(X, state)
        output = self.output_layer.dense(Y)
        return output, state

    def begin_state(self, batch_size, ctx, *args, **kwargs):
        # 1 * batch_size * hidden_size
        return self.rnn_layer.rnn.begin_state(batch_size=batch_size, ctx=ctx, *args, **kwargs)

    def detach(self, hidden):
        if isinstance(hidden, (tuple, list)):
            hidden = [self.detach(h) for h in hidden]
        else:
            hidden = hidden.detach()
        return hidden

    def _init_rnn_layer(self, config: SimpleNamespace):

        if config.mode == 'lstm':
            mode = gluon.rnn.LSTM
        elif config.mode == 'gru':
            mode = gluon.rnn.GRU
        else:
            mode = gluon.rnn.RNN

        layer = SimpleNamespace(
            rnn=mode(
                hidden_size=config.hidden_size,
                num_layers=config.num_layers,
                layout=config.layout,
                dropout=config.dropout,
                bidirectional=config.bidirectional,
                i2h_weight_initializer=config.i2h_weight_initializer,
                h2h_weight_initializer=config.h2h_weight_initializer,
                i2h_bias_initializer=config.i2h_bias_initializer,
                h2h_bias_initializer=config.h2h_bias_initializer,
                input_size=config.input_size,
            ),
            clip=config.clip
        )

        with self.name_scope():
            self.__setattr__(layer.rnn.name, layer.rnn)

        return layer

    def _init_output_layer(self, config: SimpleNamespace) -> SimpleNamespace:
        """
        :param config: the output of :meth:`RNNModel.output_layer`.
        :return: the namespace of (dense) for the output layer.
        """
        layer = SimpleNamespace(
            dense=gluon.nn.Dense(
                units=config.num_class,
                flatten=config.flatten
            )
        )

        with self.name_scope():
            self.__setattr__(layer.dense.name, layer.dense)

        return layer

